/*++

Copyright (c) 2009
Lysann Kessler, Stephanie Platz, Thomas Klingbeil, Philipp Tessenow, Frank Schlegel;
Michael Haupt, Michael Perscheid, Robert Hirschfeld;
Software Architecture Group, Hasso-Plattner-Institut, University of Potsdam, Germany

This software is published under MIT license.
It uses phidget library by Phidgets Inc. (http://phidgets.com).
See ../../../../license.pdf for more information.

Module Name:

    Phidget.c

Abstract:

    This file contains Phidget general static methods and virtual methods
	common to all Phidgets.

--*/

#include "Phidget.h"

//////////////////////////////////////////////////////////////////////////
// forward declarations: callback functions

int CALLBACK phidget_attachHandler(CPhidgetHandle handle, void *pUserPtr);
int CALLBACK phidget_detachHandler(CPhidgetHandle handle, void *pUserPtr);
int CALLBACK phidget_errorHandler(CPhidgetHandle handle, void *pUserPtr, int iErrorCode, const char *pcErrorString);

//////////////////////////////////////////////////////////////////////////
// static methods

/*++

Routine Description:

	This static method returns the last error code caught by any Phidget's
	error handler. This value is is shared in a global variable accessable
	for all Phidgets managed by the plugin.
	If no error occured or no Phidget existed when the error occured, the
	return value is 0.

Arguments:

	None.

Return Value:

	An integer identifying the last error code.
	See the official Phidget documentation for more information.

--*/
int phidget_LastError()
{
	return g_iLastError;
}

//////////////////////////////////////////////////////////////////////////
// Phidget methods

/*++

Routine Description:

	This method returns a Phidget's device class.

Arguments:

	pPhidget - (this-)pointer to the Phidget whose class to be identified.

Return Value:

	A CPhidget_DeviceClass identifying the Phidget's device class.
	See the official Phidget documentation for more information.

--*/
CPhidget_DeviceClass phidget_GetDeviceClass(Phidget *pPhidget)
{
	if(pPhidget == NULL || pPhidget->deviceClass == 0)
		return PHIDCLASS_NOTHING;
	else return pPhidget->deviceClass;
}

/*++

Routine Description:

	This method is called from all specific phidgetmodules while creating
	a phidget struct. To identify the phidgets in Squeak they need an id,
	which is set here. The VM will get a ByteArray (32 or 64 Bit) to
	represent the id.
	The created phidget will be added to the global phidget list to have
	access to all created phidget objects.

Arguments:

	pPhidget - pointer to the Phidget to be assinged the id.

Return Value:

	Squeak VM identifier for the PhigdetId Object.

--*/
sqInt phidget_Create(Phidget* pPhidget)
{
	sqInt result = 0;

	if(pPhidget == NULL)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Null pointer given in phidget_Create for pPhidget."));
		return -1;
	}

	// add to global Phidgets list
	pPhidget->id = phidgetPlugin_getNextPhidgetId();
	phidgetPlugin_AddPhidget(pPhidget);

	// return the id
	result = interpreterProxy->instantiateClassindexableSize(
		interpreterProxy->classByteArray(), sizeof(PhidgetId));
	*((PhidgetId*)(result + 4)) = pPhidget->id;

	IF_DBG(SLog(SLOG_LEVEL_INFO, "Phidget created, id = %i", pPhidget->id));
	return result;
}

//
//reset common semaphores
//
void phidget_ResetSemaphores(Phidget *pPhidget)
{
	if(pPhidget == NULL)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Null pointer given in phidget_ResetSemaphores for pPhidget."));
		return;
	}

	pPhidget->iSemaphoreAttach = 0;
	pPhidget->iSemaphoreDetach = 0;
	pPhidget->iSemaphoreError = 0;
}

/*++

Routine Description:

	This method opens a previously created Phidget with the given serial
	number.

Arguments:

	id - id of the Phidget to open the hardware connection.

	iSerialNumber - The serial number of the Phidget to be opened.
		Every Phidget has an unique serial number. Supply this here to
		open a specific Phidget.
		Set this parameter to -1 to open any Phidget with matching device
		class (supplied when creating the Phidget object).

Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_Open(PhidgetId id, int iSerialNumber)
{
	BOOL result = FALSE;
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	// do specific open operations for the phidget
	result = phidget_OpenSpecific(pPhidget);
	if(result)
	{
		// open the phidget for device connections
		result = phidget_OpenConnection_Local(pPhidget, iSerialNumber);
	}

	if(!result)
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Could not open Phidget, id = %i, serial number = %i", id, iSerialNumber));
	return result;
}

/*++

Routine Description:

	This method calls the specific open operation for a given phidget. In
	that method the actual phidget object is created.

Arguments:

	pPhidget - pointer of the Phidget to open the hardware connection.
	
Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_OpenSpecific(Phidget* pPhidget)
{
	switch(phidget_GetDeviceClass(pPhidget))
	{
	case PHIDCLASS_INTERFACEKIT:
		return phidgetIfKit_Open((PhidgetIfKit*)pPhidget);

	case PHIDCLASS_SERVO:
		return phidgetServo_Open((PhidgetServo*)pPhidget);

	case PHIDCLASS_RFID:
		return phidgetRFID_Open((PhidgetRFID*)pPhidget);

	case PHIDCLASS_ACCELEROMETER:
		return  phidgetAccel_Open((PhidgetAccel*)pPhidget);

	case PHIDCLASS_MOTORCONTROL:
		return phidgetMotCon_Open((PhidgetMotCon*)pPhidget);

	case PHIDCLASS_ADVANCEDSERVO:
		return phidgetAdvServo_Open((PhidgetAdvServo*)pPhidget);

	case PHIDCLASS_TEXTLCD:
		return phidgetTextLCD_Open((PhidgetTextLCD*)pPhidget);
	
	default:
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Could not open Phidget (specific)\n%s", phidgetPlugin_GetPhidgetDescription(pPhidget)));
		return FALSE;
	}
}

/*++

Routine Description:

	This method is called from all specific phidgetmodules while opening a
	phdiget object.
	This method registers the callbacks that will run if an attachment,
	detachment or an error occurs.

Arguments:

	pPhidget - pointer to the Phidget to register the callbacks.

Return Value:

	TRUE if the pointer is valid, FALSE otherwise.

--*/
BOOL phidget_Open_Base(Phidget* pPhidget)
{
	if(pPhidget == NULL)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Null pointer given in phidget_Open_Base for pPhidget."));
		return FALSE;
	}

	//
	// Set the handlers to be run when the device is plugged in or opened from software, unplugged or closed
	//from software, or generates an error.
	//
	CPhidget_set_OnAttach_Handler((CPhidgetHandle)(pPhidget->handle), phidget_attachHandler, (void *)pPhidget->id);
	CPhidget_set_OnDetach_Handler((CPhidgetHandle)(pPhidget->handle), phidget_detachHandler, (void *)pPhidget->id);
	CPhidget_set_OnError_Handler((CPhidgetHandle)(pPhidget->handle), phidget_errorHandler, (void *)pPhidget->id);

	return TRUE;
}

/*++

Routine Description:

	This method finally opens a device connection with the (previously
	created) Phidget with the given serial number.

Arguments:

	pPhidget - pointer of the Phidget to open the hardware connection.

	iSerialNumber - The serial number of the Phidget to be opened.

Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_OpenConnection_Local(Phidget* pPhidget, int iSerialNumber)
{
	BOOL result;

	// open the connection
	result = CPhidget_open(pPhidget->handle, iSerialNumber) == 0;

	// log
	if(result)
		IF_DBG(SLog(SLOG_LEVEL_INFO, "Phidget opened locally, serial number = %i\n%s",
			iSerialNumber, phidgetPlugin_GetPhidgetDescription(pPhidget)));
	else
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Could not open locally, serial number = %i, PhidgetId = %i\n",
			iSerialNumber, pPhidget->id));

	return result;
}

/*++

Routine Description:

	This method closes a (previously created) Phidget with the given
	id and frees its handle. Additionally reset the callbacks.

Arguments:

	id - id of the Phidget to close the hardware connection.

Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_Close(PhidgetId id)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	if(pPhidget == NULL)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Null pointer given in phidget_Close for pPhidget"));
		return FALSE;
	}
	else if(pPhidget->handle == NULL)
	{
		IF_DBG(SLog(SLOG_LEVEL_WARNING,
			"Trying to close Phidget (id = %i), but it is not opened.", id));
		return FALSE;
	}

	// close the phidget and delete the given handle
	CPhidget_close((CPhidgetHandle)(pPhidget->handle));
	CPhidget_delete((CPhidgetHandle)(pPhidget->handle));
	pPhidget->handle = 0;
	IF_DBG(SLog(SLOG_LEVEL_INFO, "Phidget closed, id = %i", pPhidget->id));

	// unregister the callbacks 
	switch(phidget_GetDeviceClass(pPhidget))
	{
	case PHIDCLASS_INTERFACEKIT:
		phidgetIfKit_ResetSemaphores((PhidgetIfKit*)pPhidget);
		return TRUE;

	case PHIDCLASS_SERVO:
		phidgetServo_ResetSemaphores((PhidgetServo*)pPhidget);
		return TRUE;

	case PHIDCLASS_RFID:
		phidgetRFID_ResetSemaphores((PhidgetRFID*)pPhidget);
		return TRUE;

	case PHIDCLASS_ACCELEROMETER:
		phidgetAccel_ResetSemaphores((PhidgetAccel*)pPhidget);
		return TRUE;

	case PHIDCLASS_MOTORCONTROL:
		phidgetMotCon_ResetSemaphores((PhidgetMotCon*)pPhidget);
		return TRUE;

	case PHIDCLASS_ADVANCEDSERVO:
		phidgetAdvServo_ResetSemaphores((PhidgetAdvServo*)pPhidget);
		return TRUE;
			
	case PHIDCLASS_TEXTLCD:
		phidgetTextLCD_ResetSemaphores((PhidgetTextLCD*)pPhidget);
		return TRUE;
			
	default:
		IF_DBG(SLog(SLOG_LEVEL_WARNING, "Unknown Phidget type %i in phidget_Close.", phidget_GetDeviceClass(pPhidget)));
		return FALSE;
	}
}

/*++

Routine Description:

	This method sets a semaphore to be signaled when the phidget with the
	corresponding id is attached.

Arguments:

	id - id of the Phidget to set the semaphore.

	iSemaphore - the semaphore the iSemaphoreAttach is set to.

Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_SetSemaphoreIndex_Attach(PhidgetId id, int iSemaphore)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	pPhidget->iSemaphoreAttach = iSemaphore;
	IF_DBG(SLog(SLOG_LEVEL_VERBOSE, "New semaphore for <Phidget-Attach> is: %d, Phidget id = %i",
		pPhidget->iSemaphoreAttach, pPhidget->id));
	return TRUE;
}

/*++

Routine Description:

	This method sets a semaphore to be signaled when the phidget with the
	corresponding id is detached.

Arguments:

	id - id of the Phidget to set the semaphore.

	iSemaphore - the semaphore the iSemaphoreDetach is set to.

Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_SetSemaphoreIndex_Detach(PhidgetId id, int iSemaphore)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	pPhidget->iSemaphoreDetach = iSemaphore;
	IF_DBG(SLog(SLOG_LEVEL_VERBOSE, "New semaphore for <Phidget-Detach> is: %d, Phidget id = %i",
		pPhidget->iSemaphoreDetach, pPhidget->id));
	return TRUE;
}

/*++

Routine Description:

	This method set a semaphore to be signaled when the phidget with the
	corresponding id causes an error.

Arguments:

	id - id of the Phidget to set the semaphore.

	iSemaphore - the semaphore the iSemaphoreError is set to.

Return Value:

	TRUE if the operation succeeded, FALSE otherwise.

--*/
BOOL phidget_SetSemaphoreIndex_Error(PhidgetId id, int iSemaphore)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	pPhidget->iSemaphoreError = iSemaphore;
	IF_DBG(SLog(SLOG_LEVEL_VERBOSE, "New semaphore for <Phidget-Error> is: %d, Phidget id = %i",
		pPhidget->iSemaphoreError, pPhidget->id));
	return TRUE;
}

/*++

Routine Description:

	This method returns the serial number of the phidget with the
	corresponding id.

Arguments:

	id - id of the Phidget to get the serial number.

Return Value:

	The Phidget's serial number.
	If there was an error, -1 is returned.

--*/
/*int phidget_getSerialNumber(PhidgetId id)
{
	int iSerialNumber = 0;
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, -1);
	
	if(CPhidget_getSerialNumber ((CPhidgetHandle)(pPhidget->handle), &iSerialNumber) != 0)
		return -1;
	else
		return iSerialNumber;
}*/

// TODO: kommentieren
BOOL phidget_getDeviceName(PhidgetId id, const char** ppcName, int* piNameSize)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	if(CPhidget_getDeviceName(pPhidget->handle, ppcName) != EPHIDGET_OK)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Cannot get name of phidget, id=%i", id));
		return FALSE;
	}

	*piNameSize = strlen(*ppcName); // without '\0'
	return TRUE;
}

// TODO: kommentieren
BOOL phidget_getSerialNumber(PhidgetId id, int* piSerialNumber)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	if(CPhidget_getSerialNumber(pPhidget->handle, piSerialNumber) != EPHIDGET_OK)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Cannot get serial number of phidget, id=%i", id));
		return FALSE;
	}

	return TRUE;
}

// TODO: kommentieren
BOOL phidget_getLabel(PhidgetId id, const char** ppcLabel, int* piLabelSize)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	if(CPhidget_getDeviceLabel(pPhidget->handle, ppcLabel) != EPHIDGET_OK)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Cannot get label of phidget, id=%i", id));
		return FALSE;
	}

	*piLabelSize = strlen(*ppcLabel); // without '\0'
	return TRUE;
}

// TODO: kommentieren
BOOL phidget_getDeviceClass(PhidgetId id, CPhidget_DeviceClass* pClass)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	if(CPhidget_getDeviceClass(pPhidget->handle, pClass) != EPHIDGET_OK)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Cannot get class of phidget, id=%i", id));
		return FALSE;
	}

	return TRUE;
}

// TODO: kommentieren
BOOL phidget_getDeviceType(PhidgetId id, CPhidget_DeviceID* pDeviceID)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	if(CPhidget_getDeviceID(pPhidget->handle, pDeviceID) != EPHIDGET_OK)
	{
		IF_DBG(SLog(SLOG_LEVEL_ERROR, "Cannot get type of phidget, id=%i", id));
		return FALSE;
	}

	return TRUE;
}

/*++

Routine Description:

	This method get the device state of a phidget with the corresponding
	id. If the phidget is attached TRUE will be returned, FALSE otherwise.

Arguments:

	id - id of the Phidget to get the attachement state.

Return Value:

	TRUE if phidget is attached, FALSE otherwise.

--*/
BOOL phidget_isAttached(PhidgetId id)
{
	BOOL bDeviceStatus;
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	CPhidget_getDeviceStatus ((CPhidgetHandle) pPhidget->handle, &bDeviceStatus);
	return bDeviceStatus; //attached=TRUE=1, not_attached=FALSE=0
}

/*++

Routine Description:

	This method checks if the phidgets' id is valid. This is when a
	phidget with this id in the global list of still created phidgets
	exists.

Arguments:

	id - id of the Phidget to get the valid state.

Return Value:

	TRUE if phidget id is valid, FALSE otherwise.

--*/
BOOL phidget_isValid(PhidgetId id)
{
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, FALSE);

	// if the last instruction didn't return with FALSE, the Phidget id is valid
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////
// callback functions

/*++

Routine Description:

	This callback is called when a phidget with the id pUserPtr is
	attached. It signals the semaphore iSemaphoreAttach.

Arguments:

	handle - handle of the Phidget to signal the semaphore.

	pUserPtr - Callback function pointer.

Return Value:

	1 if the operation succeeded, 0 otherwise.

--*/
int CALLBACK phidget_attachHandler(CPhidgetHandle handle, void *pUserPtr)
{
	PhidgetId id = (PhidgetId) pUserPtr;
	int i;
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, 0);

	IF_DBG(SLog(SLOG_LEVEL_VERBOSE, "<Phidget-Attach> id = %i, semaphore = %i",
		pPhidget->id, pPhidget->iSemaphoreAttach));

	//
	//set the Accelerometer sensitivity to 0.05 so that there will be no 
	//change, if accelerometer isn't in move
	//
	switch (phidget_GetDeviceClass(pPhidget))
	{
		case PHIDCLASS_ACCELEROMETER:
			
			for (i=0; i < phidgetCompAxis_GetCount(id); i++)
				phidgetComponent_SetIndexedDoubleChangeTrigger(id, i, 0.05 );
			break;
			
		case PHIDCLASS_TEXTLCD:
			
			for (i = 8; i < 16; i++)
				phidgetCompTextLCD_SetCustomChar(id, i, 0, 0);
			break;
	}

	if (pPhidget->iSemaphoreAttach > 0)
		interpreterProxy->signalSemaphoreWithIndex(pPhidget->iSemaphoreAttach);
	else
		IF_DBG(SLog(SLOG_LEVEL_WARNING, "No semaphore set for <Phidget-Attach>"));
	return 0;
}

/*++

Routine Description:

	This callback is called when a phidget with the id pUserPtr is
	detached. It signals the semaphore iSemaphoreDetach.

Arguments:

	handle - handle of the Phidget to signal the semaphore.

	pUserPtr - Callback function pointer.

Return Value:

	1 if the operation succeeded, 0 otherwise.

--*/
int CALLBACK phidget_detachHandler(CPhidgetHandle handle, void *pUserPtr)
{
	PhidgetId id = (PhidgetId) pUserPtr;
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, 0);

	IF_DBG(SLog(SLOG_LEVEL_VERBOSE, "<Phidget-Detach> id = %i, semaphore = %i",
		pPhidget->id, pPhidget->iSemaphoreDetach));

	if (pPhidget->iSemaphoreDetach > 0)
		interpreterProxy->signalSemaphoreWithIndex(pPhidget->iSemaphoreDetach);
	else
		IF_DBG(SLog(SLOG_LEVEL_WARNING, "No semaphore set for <Phidget-Detach>"));
	return 0;
}

/*++

Routine Description:

	This callback is called when a phidget with the id pUserPtr causes an
	error. It signals the semaphore iSemaphoreError.

Arguments:

	handle - Handle of the Phidget to signal the semaphore.

	pUserPtr - Callback function pointer.

	iErrorCode - Error code provided by the Phidget21 API. See the Phidget21
		Documentation for more details.

	pcErrorString - A string describing the error that occurred.

Return Value:

	1 if the operation succeeded, 0 otherwise.

--*/
int CALLBACK phidget_errorHandler(CPhidgetHandle handle, void *pUserPtr, int iErrorCode, const char *pcErrorString)
{
	PhidgetId id = (PhidgetId) pUserPtr;
	Phidget* pPhidget = phidgetPlugin_GetPhidgetFromId_ElseReturn(id, pPhidget, 0);

	IF_DBG(SLog(SLOG_LEVEL_VERBOSE, "<Phidget-Error> id = %i, semaphore = %i, code = %i, string = %s",
		pPhidget->id, pPhidget->iSemaphoreError, iErrorCode, pcErrorString));

	//give global knowledge of the last error
	g_iLastError = iErrorCode;

	if (pPhidget->iSemaphoreError > 0)
		interpreterProxy->signalSemaphoreWithIndex(pPhidget->iSemaphoreError);
	else
		IF_DBG(SLog(SLOG_LEVEL_WARNING, "No semaphore set for <Phidget-Error>"));
	return 0;
}
